#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include "fsactions.h"
#include "fusestructs.h"

// Adds a file to the filesystem
void AddFileToFS(FileSystem* fs, char* fname)
{
    int inputfd = open(fname, O_RDONLY);
    if (inputfd == -1)
    {
        fprintf(stderr, "Failed opening input file\n");
        exit(1);
    }
    struct stat tempstat;
    ino_t currentInode = 0;
    ino_t newInode;
    int newDirectBlock;
    char* tempname;
    stat(fname, &tempstat);
    // Get first directory, labelled by 0 to splitLocation-1,
    //      assuming splitLocation not 0;
    int splitLocation = FindNextDirectory(fname);
    if (splitLocation == 0)
    {
        fname = fname + 1; 
        splitLocation = FindNextDirectory(fname);
    }
    // While there is a '/', make a new inode out of left and keep right side
    while (splitLocation > -1)
    {
        tempname = strndup(fname, splitLocation);
        newInode = FindDirectory(fs->inodes[currentInode], tempname);
        if (newInode == -1)
        {
            newDirectBlock = GetFreeBlockNumber(fs->fbl.freeList + (currentInode*4), 4);
            newInode = GetFreeInodeNumber(fs->inodes);
            strcpy(fs->inodes[currentInode].blocks[newDirectBlock].name, tempname);
            free(tempname);
            SetupDirectoryBlock(&(fs->inodes[currentInode].blocks[newDirectBlock]), newInode);
            fs->inodes[newInode].isValid = 1;
        }
        currentInode = newInode;
        fname = fname + splitLocation + 1;
        splitLocation = FindNextDirectory(fname);
    }
    newDirectBlock = GetFreeBlockNumber(fs->fbl.freeList + (currentInode*4), 4);
    newInode = currentInode*4 + newDirectBlock;
    fs->inodes[currentInode].blocks[newDirectBlock].isValid = 1;
    fs->inodes[currentInode].blocks[newDirectBlock].size = tempstat.st_size;
    strcpy(fs->inodes[currentInode].blocks[newDirectBlock].name, fname);
    fs->inodes[currentInode].blocks[newDirectBlock].inode = newInode;
    read(inputfd, &(fs->dataBlocks[newInode].byte), tempstat.st_size);
    close(inputfd);
    return;
}

// List path names of all used inodes
void ListFS(FileSystem* fs)
{
    printf("/\n");
    _ListFS_(fs, 0, 1);
    return;
}

void _ListFS_(FileSystem* fs, ino_t inodeNumber, short unsigned numtabs)
{
    for (int i = 0; i < fs->superBlock.inodeCount; i++)
    {
        if (!(fs->inodes[inodeNumber].blocks[i].isValid))
            continue;
        for (int j = 0; j < numtabs; j++)
            printf("\t");
        printf("%s\n", fs->inodes[inodeNumber].blocks[i].name);
        if (fs->inodes[inodeNumber].blocks[i].isDirectory)
            _ListFS_(fs, fs->inodes[inodeNumber].blocks[i].inode, numtabs + 1);
    }
}

// Remove a file from the filesystem
void RemoveFileFromFS(FileSystem* fs, char* fname)
{
    _RemoveFileFromFS_(fs, 0, fname);
}

// Recursive bool call for checking before removing
// If the directory is not empty, returns false
int _RemoveFileFromFS_(FileSystem* fs, ino_t inodeNumber, char* fname)
{
    int emptyDir = 1;
    for (int i = 0; i < DEFAULTINODEMAX; i++)
    {
        if (!(fs->inodes[inodeNumber].blocks[i].isValid))
            continue;
        if (fs->inodes[inodeNumber].blocks[i].isDirectory)
        {
            emptyDir =_RemoveFileFromFS_(fs, fs->inodes[inodeNumber].blocks[i].inode, fname + FindNextDirectory(fname) + 1);
            if (emptyDir)
                fs->inodes[inodeNumber].blocks[i].isValid = 0;
            continue;
        }
        if (strcmp(fs->inodes[inodeNumber].blocks[i].name, fname) == 0)
            fs->inodes[inodeNumber].blocks[i].isValid = 0;
        else
            emptyDir = 0;
    } 
    if (!emptyDir)
        return 0;
    fs->inodes[inodeNumber].isValid = 0;
    return 1;
}

// Extract a file from the filesystem
void ExtractFileFromFS(FileSystem* fs, char* fname)
{
    
    ino_t currentInode = 0;
    ino_t newInode;
    char* tempname;
    // Get first directory, labelled by 0 to splitLocation-1,
    //      assuming splitLocation not 0;
    int splitLocation = FindNextDirectory(fname);
    if (splitLocation == 0)
    {
        fname = fname + 1; 
        splitLocation = FindNextDirectory(fname);
    }
    // While there is a '/', make a new inode out of left and keep right side
    while (splitLocation > -1)
    {
        tempname = strndup(fname, splitLocation);
        newInode = FindDirectory(fs->inodes[currentInode], tempname);
        if (newInode == -1)
        {
            fprintf(stderr, "Error extracting file due to path\n");
            exit(1);
        }
        currentInode = newInode;
        fname = fname + splitLocation + 1;
        splitLocation = FindNextDirectory(fname);
    }
    for (int i = 0; i < DEFAULTINODEMAX; i++)
    {
        if (strcmp(fs->inodes[currentInode].blocks[i].name, fname) == 0)
        {
            newInode = fs->inodes[currentInode].blocks[i].inode;
            off_t size = fs->inodes[currentInode].blocks[i].size;
            write(STDOUT_FILENO, fs->dataBlocks[newInode].byte, size);
            return;
        }
    }
    return;
}


// Get the next free inode number
ino_t GetFreeInodeNumber(Inode inodes[])
{
    for (int i = 0; i < DEFAULTINODEMAX; i++)
        if (!(inodes[i].isValid))
            return i;
    return -1;
}

// Gets next free inode number and returns it
ino_t GetFreeBlockNumber(int fbl[], unsigned short size)
{
    int bitPosition;
    for (int i = 0; i < size; i++)
    {
        bitPosition = FindEmptyBitPosition(fbl[i]);
        if (bitPosition >= 0)
        {
            fbl[i] = (fbl[i] | (1 << bitPosition));
            return ((i*32) + bitPosition);
        }
    }
    return -1;
}

// Finds next empty bit in Free Block List
int FindEmptyBitPosition(int number)
{
    for (int i = 0; i < 32; i++)
        if ((~number & (1 << i)) == (1 << i))
            return i;
    return -1;
}

// Finds the next split character in the path
int FindNextDirectory(char* fname)
{
    int i = 0;
    while (fname[i] != '\0')
    {
        if (fname[i] == '/')
            return i;
        i++;
    }
    return -1;
}

// Function checks if directory is setup. If it is, then returns the inode
int FindDirectory(Inode inode, char* directoryName)
{
    for (int i = 0; i < 128; i++)
    {
        if (!(inode.blocks[i].isValid))
            continue;
        if (!(inode.blocks[i].isDirectory))
            continue;
        if (strcmp(inode.blocks[i].name, directoryName) == 0)
            return inode.blocks[i].inode;
    }
    return -1;
}

// Function for setting all defaults of filesystem
void SetFileSystemDefaults(FileSystem* fs)
{
    fs->superBlock.blockSize = DEFAULTBLOCKSIZE;
    fs->superBlock.blockCount = DEFAULTBLOCKMAX;
    fs->superBlock.inodeCount = DEFAULTINODEMAX;
    for (int i = 0; i < 4; i++)
        fs->fbl.freeList[i] = 0;
    for (int i = 0; i < DEFAULTINODEMAX; i++)
        fs->inodes[i].isValid = 0;
    return;
}

// Sets first inode as root directory
void SetupRootDirectory(FileSystem* fs)
{
    fs->inodes[0].isValid = 1;
    return;
}

void SetupDirectoryBlock(DirectBlock* newBlock, ino_t inode)
{
    newBlock->isValid = 1;
    newBlock->isDirectory = 1;
    newBlock->inode = inode;
    return;
}
